/*
 * MatrixProductPiece.java
 *
 */

package Application;

import dku.DKUPiece;

/** This Version of Matrix Multiply is only intended for shared memory
 *   machines.  A distributed memory version must include code that physically
 *   divides the original matrices, rather than only dividing the iteration
 *   space.
 *
 * @author Me
 */
public class MatrixProductPiece implements DKUPiece
 {
   // pointers to original data..  the result matrix must be created when the
   //  left and right matrices are put into the original, root, piece-object.
   RowMajorMatrix      leftMatrix;
   ColMajorMatrix      rightMatrix;  //making ColMajor in anticipation of distr
                                     // memory version
   RowMajorMatrix      resultMatrix;
      
   // define the starting and ending boundaries for this piece of the result
   //  matrix.  These are derivable from the left and right matrices.. included
   //  them explicitly for readability of code.
   int prodStartRow, prodEndRow;
   int prodStartCol, prodEndCol;
   // Start and end of the portion of the left matrix that contributes to this
   //  piece of the product
   int leftStartRow, leftEndRow;
   int leftStartCol, leftEndCol;
   // Start and end of the portion of the right matrix that contributes to this
   //  piece of the product
   int rightStartRow, rightEndRow;
   int rightStartCol, rightEndCol;
   
   // Part of pattern.  These should appear in every DKUPiece implementor
   MatrixProductPiece[]  subPiecesArray;
   int            numSubPieces = 0;
   int            currentSubPieceInIter = 0;
   int            numSubPiecesFinished = 0;
   boolean        moreSubPiecesToUndivide = true;
   
   
   /** Use this constructor to make a root piece..  it's only for a root piece
    * 
    * @param _leftMatrix
    * @param _rightMatrix
    */
   public MatrixProductPiece( RowMajorMatrix _leftMatrix,
                              ColMajorMatrix _rightMatrix )
    { 
      leftMatrix   = _leftMatrix;
      rightMatrix  = _rightMatrix;
      resultMatrix = 
          new RowMajorMatrix( leftMatrix.numRows, rightMatrix.numCols );
      leftStartRow  = 0;
      leftEndRow    = leftMatrix.numRows - 1;
      leftStartCol  = 0;
      leftEndCol    = leftMatrix.numCols - 1;

      rightStartRow = 0;
      rightEndRow   = rightMatrix.numRows - 1;
      rightStartCol = 0;
      rightEndCol   = rightMatrix.numCols - 1;

      prodStartRow  = leftStartRow;
      prodEndRow    = leftEndRow;
      prodStartCol  = rightStartCol;
      prodEndCol    = rightEndCol;
    }
   public MatrixProductPiece( MatrixProductPiece incomingPiece )
    { 
      leftMatrix   = incomingPiece.leftMatrix;
      rightMatrix  = incomingPiece.rightMatrix;
      resultMatrix = incomingPiece.resultMatrix;
    }

   /** These two iterator methods are part of the pattern.  It is an iterator,
    *   however the
    *   semantics of the pattern state that the sub-pieces can be gotten in a
    *   non-deterministic order, and processed in kernels with undefined overlap
    */
   public boolean hasMoreSubPieces()
    { return currentSubPieceInIter < numSubPieces;
    }
   public DKUPiece giveNextSubPiece()
    { DKUPiece
      retPiece = subPiecesArray[ currentSubPieceInIter ];
      currentSubPieceInIter += 1;
      return retPiece;
    }
   
    
   /** Replaced by hardware-specific code..  a separate variable keeps the
    *   boolean value in order to make translation easier.
    */
   public boolean  hasMoreSubPiecesToUndivide()
    { return moreSubPiecesToUndivide;
    }
   
   public void reset()
    { 
      leftMatrix = null;
      rightMatrix = null;
      resultMatrix = null;
   
      subPiecesArray = null;
      numSubPieces = 0;
      currentSubPieceInIter = 0;
      numSubPiecesFinished = 0;
      moreSubPiecesToUndivide = true;
    }
   
   public void resetIter()
    { currentSubPieceInIter = 0;
    }
   
   
   /** Dividing the iteration space..  the data structures get divided into
    *   whatever each sub-iteration-space needs
    * 
    *  In the case of matrix multiply, dividing a product-piece into, for
    *   example, 4, means 
    *   dividing the iteration space that product-piece represents into 4,
    *   which means dividing the left matrix by 2, and the right matrix by 2,
    *   and talking all pairings of a left piece times a right piece.
    * 
    * @param numPieces
    */
   public void divideSelfInto_SubPieces( int numPieces )
    {
      // need numPieces to be divisible by two integers
      // but, for now, being lazy, so require it to be a perfect square
      // So, if it's not, make it into the next larger perfect square
      double foo = Math.sqrt( numPieces );
      int numMatrixPieces = (int) Math.ceil( foo );
      numSubPieces = numMatrixPieces * numMatrixPieces;
      
      // Gets a bit confusing..  Matrix Product Piece is a piece of iteration
      //  space..  it is the iterations in which one piece of the left matrix
      //  is multiplied by one piece of the right matrix.
      // Thus, to make product pieces, both the left and right matrices have
      //  to be divided, then all pairings of those pieces taken.  Each \
      //  pairing on matrix pieces is one product piece.
      // Notice that the number of product pieces is numMatrixPieces squared.
      int numLeftRows, numLeftCols, numRightRows, numRightCols;
      numLeftRows  = leftEndRow - leftStartRow + 1; // +1 cause starts at zero
      numLeftCols  = leftEndCol - leftStartCol + 1;
          
      numRightRows = rightEndRow - rightStartRow + 1;
      numRightCols = rightEndCol - rightStartCol + 1;
      
      // This implementation is only for shared memory, so won't divide up the
      //  actual matrices, just make ranges of indexes.
      // Make a sub-matrix be represented by a restriction on the start
      //  and end rows that are saved inside the product-piece
      if( numMatrixPieces > 1 && numLeftRows > 1 )
       {    // sched wants more than one piece, and matrices big enough to div.
                  
            // this division drops the fractional part, like floor()
         int
         rowsPerLeftPiece = (int) ( numLeftRows / numMatrixPieces );
         if( rowsPerLeftPiece == 0 ) // check size of pieces
          { rowsPerLeftPiece = 1;    // if too small, make it min sz
            numMatrixPieces = numLeftRows; 
          }
         int
         colsPerRightPiece = rowsPerLeftPiece; //left rows==right cols
         
         // numMatrixPieces may have changed so re-calculate numSubPieces
         numSubPieces = numMatrixPieces * numMatrixPieces;
         subPiecesArray = new MatrixProductPiece[ numSubPieces ];
         
         // make the matrix pieces (divide up the iteration space)
         MatrixIterSpacePiece newLeftPiece, newRightPiece;
         MatrixIterSpacePiece[] leftPieces, rightPieces;
         leftPieces = new MatrixIterSpacePiece[ numMatrixPieces ];
         rightPieces = new MatrixIterSpacePiece[ numMatrixPieces ];
         
            // make one fewer matrix piece..  the last may be odd-sized
         for(int matrixPieceNum = 0; matrixPieceNum < numMatrixPieces - 1;
                 matrixPieceNum += 1 )
          { newLeftPiece  = new MatrixIterSpacePiece();
            newRightPiece = new MatrixIterSpacePiece();
                                     
            newLeftPiece.startRow = matrixPieceNum * rowsPerLeftPiece;
            newLeftPiece.endRow   = newLeftPiece.startRow + rowsPerLeftPiece;
            newLeftPiece.startCol = 0;
            newLeftPiece.endCol   = numLeftCols - 1;
            
            newRightPiece.startRow = 0;
            newRightPiece.endRow   = numRightRows - 1;
            newRightPiece.startCol = matrixPieceNum * colsPerRightPiece;
            newRightPiece.endCol   = newRightPiece.startCol + colsPerRightPiece;
                                    
            leftPieces [ matrixPieceNum ] = newLeftPiece;
            rightPieces[ matrixPieceNum ] = newRightPiece;
          }
            // make the last, possibly odd-sized, piece
         newLeftPiece  = new MatrixIterSpacePiece();
         newRightPiece = new MatrixIterSpacePiece();

         newLeftPiece.startRow = (numMatrixPieces - 1) * rowsPerLeftPiece;
         newLeftPiece.endRow   = numLeftRows - 1;
         newLeftPiece.startCol = 0;
         newLeftPiece.endCol   = numLeftCols - 1;

         newRightPiece.startRow = 0;
         newRightPiece.endRow   = numRightRows - 1;
         newRightPiece.startCol = (numMatrixPieces - 1) * colsPerRightPiece;
         newRightPiece.endCol   = numRightCols - 1;

         leftPieces [ numMatrixPieces - 1 ] = newLeftPiece;
         rightPieces[ numMatrixPieces - 1 ] = newRightPiece;
            
                                  
         // pair up the pieces
         int newPiecePos;
         MatrixProductPiece newProdPiece;
         MatrixIterSpacePiece leftMPiece, rightMPiece;
         for(int rightMatrixPieceNum = 0; rightMatrixPieceNum < numMatrixPieces;
                 rightMatrixPieceNum +=1 )
          { for(int leftMatrixPieceNum =0; leftMatrixPieceNum < numMatrixPieces;
                    leftMatrixPieceNum +=1 )
             { 
               newProdPiece = new MatrixProductPiece( this );  //copies pointers
               
               leftMPiece  = leftPieces [ leftMatrixPieceNum  ];
               rightMPiece = rightPieces[ rightMatrixPieceNum ];
               
               newProdPiece.leftStartRow  = leftMPiece.startRow;
               newProdPiece.leftEndRow    = leftMPiece.endRow;
               newProdPiece.leftStartCol  = leftMPiece.startCol;
               newProdPiece.leftEndCol    = leftMPiece.endCol;
               
               newProdPiece.rightStartRow  = rightMPiece.startRow;
               newProdPiece.rightEndRow    = rightMPiece.endRow;
               newProdPiece.rightStartCol  = rightMPiece.startCol;
               newProdPiece.rightEndCol    = rightMPiece.endCol;
               
               newProdPiece.prodStartRow  = leftMPiece.startRow;
               newProdPiece.prodEndRow    = leftMPiece.endRow;
               newProdPiece.prodStartCol  = rightMPiece.startCol;
               newProdPiece.prodEndCol    = rightMPiece.endCol;
               
               newPiecePos = rightMatrixPieceNum * numMatrixPieces + 
                             leftMatrixPieceNum; 
               subPiecesArray[ newPiecePos ] = newProdPiece;
             }
          }//for(int rightMatrixPieceNum = 0; rightMatrixPieceNum < 
         
       }//if( numMatrixPieces > 1 && numLeftRows > 1 )
      
    }
   
   /** Produces the portion of the result that is calculated in the portion of
    *   the iteration space that is in this object's piece of the iteration
    *   space.
    *  Have start and end rows and cols for both left matrix and right matrix,
    *   so,\ this is a loop that performs the matrix product.  Because only
    *   slicing the matrices, a full vector is always available inside this
    *   loop.  That means, that this loop next generates the final results,
    *   they do not need to be combined with partial results from other kernels.
    *  (by "slicing" mean only split left into rows and right into cols..  the
    *   number of cols in all left-pieces is same as cols in orinal matrix,
    *   and equivalently for rows in right matrix.)
    */
   public void performKernelOnSelf()
    { 
      for( int row = leftStartRow; row <= leftEndRow; row++ )
       { for( int col = rightStartCol; col <= rightEndCol; col++ )
          { float sum = 0;
            int
            vectorSize = leftMatrix.numCols;
            for( int i = 0; i < vectorSize; i++)
             { sum += leftMatrix.rows[row][ i ] * rightMatrix.cols[ col ][ i ];
             }
            resultMatrix.rows[ row ][ col ] = sum;
          }
       }
    }
    
   
   /**
    * Scheduler calls this method, handing it one of the sub-pieces made in
    *  this instance's divider.  Thus, a MatrixPairPiece_old instance is responsible
    *  for keeping track of how many sub-pieces are made from it, and for
    *  tracking when all of the sub-pieces have been un-divided.
    * 
    * For now, every DKUPiece is required to define several standard variables.
    *  Among those are:
    *  numSubPiecesFinished
    *  moreSubPiecesToUndivide
    * 
    * These variables have to be set inside the incrementNumSubPiecesUndivided
    *  method.
    * 
    */
   public void unDivideASubPiece( DKUPiece aFinishedSubPiece )
    {
      incrementNumSubPiecesUndivided();
    }
   
   /**
    * A hardware-specific implementation of the back-end compiler (or pre-
    *  processor) will replace this method with something hardware specific.
    *  For example, a distributed memory pre-processor might turn the method
    *  into a memory barrier implementation.  A separate boolean variable is
    *  kept in order to make translation easier.
    */
   protected void incrementNumSubPiecesUndivided()
    {
      numSubPiecesFinished += 1;  //HW-specific impl does synchronizations
      if( numSubPiecesFinished >= numSubPieces ) 
       { moreSubPiecesToUndivide = false;  // standard pattern.
       }
    }
 }   
   